from pyanaconda.flags import flags
from pyanaconda.iutil import execInSysroot, execWithRedirect, getSysroot, execWithCapture
from pyanaconda.product import productVersion
from pyanaconda import network
from pyanaconda import hvt_checker
from pyanaconda.ui.lib.disks import getDisks
import storage_utils
import logging
log = logging.getLogger("anaconda")
from blivet.errors import PartitioningError, NotEnoughFreeSpaceError
from blivet.size import Size
from blivet.devicelibs import swap
from blivet.platform import platform
from blivet.partitioning import growLVM, doPartitioning
from blivet.formats import getFormat

vstorage_bin = "/usr/bin/vstorage"

import re
import os
import time

class MDdev(object):
    def __init__(self, storage, disks, fmt_type, size = None, mountpoint = None,
            weight = 0, level = 1, grow = False):
        self.level = level
        self.grow = grow
        self.storage = storage
        self.disks = disks
        self.size = size
        self.fmt_type = fmt_type
        self.mountpoint = mountpoint
        self.weight = weight
        self.parts = []
        self.device = None
        if not disks:
            raise PartitioningError("No disks available")

    def create_members(self):
        # Fallback to basic partition when one drive available
        if len(self.disks) == 1:
            self.device = self.storage.newPartition(fmt_type=self.fmt_type,
                                            size=self.size,
                                            mountpoint=self.mountpoint,
                                            parents=self.disks,
                                            weight=self.weight,
                                            grow=self.grow)
            self.storage.createDevice(self.device)
            return
        for disk in self.disks:
            part = self.storage.newPartition(fmt_type="mdmember",
                                            size=self.size,
                                            parents=[disk],
                                            weight=self.weight,
                                            grow=self.grow)
            self.storage.createDevice(part)
            self.parts.append(part)
            disk.availsize -= self.size

    def create_md(self):
        if len(self.disks) == 1:
            return
        self.device = self.storage.newMDArray(fmt_type=self.fmt_type,
                                            memberDevices=len(self.parts),
                                            totalDevices=len(self.parts),
                                            level=self.level,
                                            mountpoint=self.mountpoint,
                                            parents=self.parts)
        self.storage.createDevice(self.device)

class DoVZAutopartPreselect(object):
    def __init__(self, instClass, storage, ksdata):
        self.instClass = instClass
        self.storage = storage
        self.ksdata = ksdata
        self.reservedSize = Size("8MiB")
        self.slash_drive = None
        self.hdds = []
        self.disks = getDisks(storage.devicetree)

    def check_min_hdd_size(self, size):
        # Function that checks for disk with required minimum size available
        return len([ d for d in self.hdds if d.availsize >= size ]) != 0

    def reset_drives_selection(self):
        for disk in self.disks:
            disk.is_ssd = disk.use_ssd = False
            disk.cap_boot = disk.use_boot = False
            disk.cap_slash = disk.use_slash = False
            disk.cap_csdata = disk.use_csdata = False
            disk.cap_vzdata = disk.use_vzdata = False

        self.preselect_drives()

    def set_csdata_non_slash_drives(self):
        # Vstorage case - vzdata on slash, hdds mark as CS used
        for disk in [ d for d in self.hdds if d.cap_csdata and
                not d.use_slash ]:
            disk.use_csdata = True
            log.info("%s drive selected for CS" % disk.name)

    def preselect_vstorage(self):
        self.set_csdata_non_slash_drives()
        # It's safe because of we skip HDDs which not capable for / + /vz
        self.slash_drive.use_vzdata = True
        log.info("%s drive selected for VZ" % self.slash_drive.name)
        self.slash_drive.availsize -= self.instClass.min_vz_size
        # Try to place CS together with slash
        if not [ d for d in self.disks if d.use_csdata ] and \
            self.slash_drive.availsize > self.instClass.min_cs_size:
                self.slash_drive.use_csdata = True
                log.info("%s slash_drive selected for CS" % self.slash_drive.name)

    def set_vzdata_non_slash_drives(self):
        # Non-vstorage - mark vzdata capable drives
        for disk in [ d for d in self.hdds if d.cap_vzdata and
                not d.use_slash ]:
            disk.use_vzdata = True
            log.info("%s drive selected for VZ" % disk.name)


    def preselect_non_vstorage(self):
        self.set_vzdata_non_slash_drives()
        if not [ d for d in self.disks if d.use_vzdata ]:
            self.slash_drive.use_vzdata = True
            log.info("%s slash_drive selected for VZ" % self.slash_drive.name)

    def preselect_drives(self):
        # First - select separatelly drives for SSD and boot
        for disk in self.disks:
            # Check for protected - and exclude such devices
            # (e.g. installation flash for now)
            # at early stage
            if disk.protected:
                log.info("Skipped protected device %s" % disk.name)
                continue

            # Also skip all other USB HDDs/flashes
            if not in_cmdline("allow_usb_hdd") and re.match(".*\/usb.*", disk.sysfsPath):
                log.info("Skipped USB device %s" % disk.name)
                continue

            # Set availsize - will use later
            # Reserve size for LVM metadata
            disk.availsize = disk.size - self.reservedSize

            # Skip not selected drives - only kickstart case
            if (self.ksdata.ignoredisk.onlyuse and not \
                    disk.name in self.ksdata.ignoredisk.onlyuse) or \
                    disk.name in self.ksdata.ignoredisk.ignoredisk:
                log.info("Skipped drive %s during autopartitioning: drive not selected" % disk.name)
                continue

            # Skip too small drives for boot
            if disk.size < self.instClass.bootreq_size:
                log.info("Skipped drive %s during autopartitioning: too small even for " \
                    "bootreq - %s" % (disk.name, disk.size))
                continue

            # Select drives for boot
            disk.availsize -= self.instClass.bootreq_size
            disk.cap_boot = disk.use_boot = True
            log.info("%s drive selected for boot" % disk.name)

            # Select drives for SSD cache
            if is_ssd(disk):
                disk.is_ssd = disk.use_ssd = True
                if disk.availsize >= self.instClass.min_hdd_size:
                    disk.cap_slash = True
                    disk.cap_vzdata = True
                if disk.availsize >= self.instClass.min_hdd_size + self.instClass.min_cs_size:
                    disk.cap_csdata = True
                log.info("%s drive selected for SSD cache" % disk.name)
                continue

            self.hdds.append(disk)

        # Try to find non-SSD drive enough to minimal swap + / + vz
        if not self.check_min_hdd_size(self.instClass.min_hdd_size):
            for disk in self.disks:
                disk.use_ssd = False
                self.hdds.append(disk)
            if not self.check_min_hdd_size(self.instClass.min_hdd_size):
                # Here we fail with autopart: BIP-BIP-BIP happens, too low space
                # Let user partition this BIP-BIP-BIP manually
                log.error("Too small size for autopart, autopart disabled")
                return

        self.storage.can_autopart = True
        log.info("Autopartition supported")

        # Select drive for / and swap
        slash_capable_hdds = [ d for d in self.hdds if d.availsize >= self.instClass.min_hdd_size ]
        for disk in slash_capable_hdds:
            disk.cap_slash = True
            log.info("%s drive is slash capable" % disk.name)
        self.slash_drive = sorted(slash_capable_hdds, key = lambda k: k.size)[0]
        self.slash_drive.use_slash = True
        log.info("%s drive selected for slash" % self.slash_drive.name)
        self.slash_drive.availsize -= (self.instClass.swap_size + self.instClass.slash_size)

        # Next, check for drives, which can be used for VZ
        for disk in [ d for d in self.hdds if d.availsize >= self.instClass.min_vz_size ]:
            disk.cap_vzdata = True
            log.info("%s drive is vzdata capable" % disk.name)

        # Next, check for drives, which can be used for CSes. Do not calculate reserved and bootreq
        # them are too small relative to min_cs_size
        for disk in [ d for d in self.hdds if d.availsize >= (self.instClass.min_cs_size - \
            self.reservedSize - self.instClass.bootreq_size) ]:
            disk.cap_csdata = True
            log.info("%s drive is csdata capable" % disk.name)

        # Next mark drives for data
        if self.storage.vstorage_cs:
            self.preselect_vstorage()
            return

        self.preselect_non_vstorage()


class DoVZAutopart(object):
    def __init__(self, instClass, storage):
        self.instClass = instClass
        self.storage = storage
        self.mddevs = []
        self.slash_drive = None
        self.vg = []
        self.csnum = 0
        self.ssdnum = 0

    @property
    def get_boot_disks(self):
        return [ d for d in self.storage.disks if d.use_boot ]

    def prepare_boot(self, boot_disks):
        from blivet.platform import EFI, MacEFI
        # Create RAID1 on every disk for boot
        # 1MiB biosboot reserve or EFI + /boot
        if isinstance(platform, EFI):
            fmt_type = "efi"
            if isinstance(platform, MacEFI):
                fmt_type = "macefi"
            self.mddevs.append(MDdev(self.storage, boot_disks, fmt_type, size=self.instClass.efiboot_size,
                mountpoint="/boot/efi", weight=2000))
            log.info("Created /boot/efi instance on %s" % ', '.join(d.name for d in boot_disks))
        else:
            # Create biosboot required by GPT on arch HDD
            for disk in [ d for d in boot_disks if getattr(d.format, "labelType", None) == "gpt" ]:
                self.storage.createDevice(self.storage.newPartition(fmt_type="biosboot",
                                                size=self.instClass.biosboot_size,
                                                parents=[disk],
                                                weight=5000))
                log.info("Created biosboot instance on %s" % disk.name)

        self.mddevs.append(MDdev(self.storage, boot_disks, "ext4", size=self.instClass.boot_size,
            mountpoint="/boot", weight=2000))
        log.info("Created /boot instance on %s" % ', '.join(d.name for d in boot_disks))

    def create_swap(self, drives):
        self.storage.createDevice(self.storage.newPartition(fmt_type="swap",
                                                size=self.instClass.swap_size,
                                                parents=drives,
                                                weight=1000))
        log.info("Created swap instance on %s" % ', '.join(d.name for d in drives))

    @property
    def vstorage_slash_max_size(self):
        return max(self.instClass.max_slash_size, self.instClass.slash_size +
            len([d for d in self.storage.disks if d.use_csdata]) * Size("1536MiB"))

    def device_func(self, device):
        device_func = self.storage.newLV
        if device.isDisk:
            device_func = self.storage.newPartition
        return device_func

    def create_slash(self, device, maxsize):
        self.storage.createDevice(self.device_func(device)(fmt_type="ext4",
                                                size=self.instClass.slash_size,
                                                mountpoint="/",
                                                parents=[device],
                                                grow=True,
                                                maxsize=maxsize))
        log.info("Created / instance on %s" % device.name)

    @property
    def drives_for_pv(self):
        return [ d for d in self.storage.disks if d.use_vzdata and not d.use_slash ]

    def create_lvm(self, pv_drives):
        # Prepare PVs
        pv_parts = []
        for disk in pv_drives:
            part = self.storage.newPartition(fmt_type="lvmpv",
                                                grow=True,
                                                parents=[disk])
            self.storage.createDevice(part)
            pv_parts.append(part)
            log.info("Created pv instance on %s" % disk.name)

        # Create LVM VG volume
        vg = self.storage.newVG(parents=pv_parts)
        self.storage.createDevice(vg)
        return vg

    def create_vz(self, drives, grow=True):
        self.storage.createDevice(self.storage.newLV(fmt_type="ext4vz",
                                                size=self.instClass.min_vz_size,
                                                mountpoint="/vz",
                                                grow=grow,
                                                parents=drives))
        log.info("Created /vz instance on %s" % ', '.join(d.name for d in drives))

    @property
    def get_slash_drives(self):
        return [ d for d in self.storage.disks if d.use_slash ]

    @property
    def get_vz_drives(self):
        return [ d for d in self.storage.disks if d.use_vzdata ]

    @property
    def get_ssd_drives(self):
        return [ d for d in self.storage.disks if d.use_ssd ]

    def create_ssd(self, device):
        self.ssdnum += 1
        self.storage.createDevice(self.storage.newPartition(fmt_type="ext4",
                                                mountpoint="/mnt/ssd%i" % self.ssdnum,
                                                size=self.instClass.min_ssd_size,
                                                grow=True,
                                                parents=[device]))
        log.info("Created SSD cache instance on %s" % device.name)

    def create_ssds(self, drives):
        [ self.create_ssd(d) for d in drives ]

    @property
    def get_cs_disks(self):
        return [ d for d in self.storage.disks if d.use_csdata and not d.use_slash ]

    def creats_cs(self, disk):
        self.csnum += 1
        self.storage.createDevice(self.device_func(disk)(fmt_type="ext4cs",
                                                mountpoint="/vstorage/%s-cs%i" %
                                                (self.storage.vstorage_cs, self.csnum),
                                                size=self.instClass.min_cs_size,
                                                grow=True,
                                                parents=[disk]))
        log.info("Created CS instance on %s" % disk.name)

    def create_vstorage_cs(self, devices):
        [ self.creats_cs(d) for d in devices ]

    @property
    def slash_together_cs(self):
        return self.slash_drive.use_csdata

    @property
    def slash_together_vz(self):
        return self.slash_drive.use_vzdata

    def create_not_vstorage_and_not_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap([self.slash_drive])
        self.vg = self.create_lvm(self.drives_for_pv)
        self.create_slash(self.slash_drive, self.slash_drive.size)
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)

    def create_not_vstorage_and_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap([self.slash_drive])
        self.vg = self.create_lvm([self.slash_drive] + self.drives_for_pv)
        self.create_slash(self.vg, self.instClass.max_slash_size)
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)

    def create_vstorage_and_not_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap([self.slash_drive])
        self.vg = self.create_lvm([self.slash_drive])
        self.create_slash(self.vg, self.vstorage_slash_max_size)
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)
        self.create_vstorage_cs(self.get_cs_disks)

    def create_vstorage_and_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap([self.slash_drive])
        self.vg = self.create_lvm([self.slash_drive])
        self.create_slash(self.vg, self.vstorage_slash_max_size)
        self.create_vz([self.vg], grow=False)
        self.create_ssds(self.get_ssd_drives)
        self.create_vstorage_cs(self.get_cs_disks)
        # And create vstorage CS on slash PV
        self.create_vstorage_cs([self.vg])

    def validate_selection(self):
        errormess = ""

        if not self.get_slash_drives:
            errormess = "At least one drive for / need to be selected.\n"
        if not self.get_vz_drives:
            errormess += "At least one drive for /vz need to be selected.\n"

        return errormess

    def log_preselect(self):
        for disk in self.storage.disks:
            log.info("Flags for %s" % disk.name)
            if disk.cap_slash:
                log.info("%s capable for slash" % disk.name)
            if disk.use_slash:
                log.info("%s selected for slash" % disk.name)
            if disk.cap_vzdata:
                log.info("%s capable for VZ" % disk.name)
            if disk.use_vzdata:
                log.info("%s selected for VZ" % disk.name)
            if disk.cap_csdata:
                log.info("%s capable for CS" % disk.name)
            if disk.use_csdata:
                log.info("%s selected for CS" % disk.name)
            if disk.is_ssd:
                log.info("%s is SSD" % disk.name)
            if disk.use_ssd:
                log.info("%s selected for SSD" % disk.name)

    def execute(self):
        if not self.storage.can_autopart:
            return

        self.log_preselect()

        # Reset drive numbers
        self.csnum = 0
        self.ssdnum = 0

        # Check selection
        errormess = self.validate_selection()
        if errormess:
            raise PartitioningError("autopart failed:\n%s" % errormess)

        self.storage.doAutoPart = True

        # GPT only!
        default_disklabel_type = "gpt"
        if not platform.setDefaultDiskLabelType(default_disklabel_type):
            raise PartitioningError("Failed to set default disktype to %s\n" %
                default_disklabel_type)

        # TODO Fill drivers to clear in GUI
        self.storage.clearPartitions()

        # Get slash
        self.slash_drive = self.get_slash_drives[0]

        # Check for OpenVZ - disable vstorage case
        if self.instClass.is_free:
            self.storage.vstorage_cs = ""

        # Will process 4 scenarios
        # Not together means that /vz placed on separate drive from /
        if not self.storage.vstorage_cs and not self.slash_together_vz:
            self.create_not_vstorage_and_not_together()
        # Together means that /vz placed on the same drive with /
        elif not self.storage.vstorage_cs and self.slash_together_vz:
            self.create_not_vstorage_and_together()
        elif self.storage.vstorage_cs and not self.slash_together_cs:
            self.create_vstorage_and_not_together()
        elif self.storage.vstorage_cs and self.slash_together_cs:
            self.create_vstorage_and_together()

        # Create members before partitions fixage to do proper size calculations
        [ md.create_members() for md in self.mddevs ]

        doPartitioning(self.storage)

        [ md.create_md() for md in self.mddevs ]

        growLVM(self.storage)

        self.storage.setUpBootLoader()

        # only newly added swaps should appear in the fstab
        self.storage.setFstabSwaps((dev for dev in self.storage.swaps if not dev.format.exists))

        errors = storage_utils.sanity_check(self.storage)
        if errors:
            raise PartitioningError("autopart failed:\n" + "\n".join(error.message
                for error in errors))

def rpm_import(src):
    """
    Import the RPM GPG key to the installed system.
    Input parameter is the full path to a key.
    """

    log.info("Import the RPM key: %s" % src)

    rc = execWithRedirect("/bin/rpm", [ "--import", src,
                          "--dbpath", getSysroot() + "/var/lib/rpm"])

    if rc:
        log.error("Failed to import the RPM key %s, error(%i)" % (src, rc))

def calculate_ver(m):
    version = 0
    if m and m.group(1) and m.group(2) and m.group(3) and m.group(4):
        version = int(m.group(1)) * pow(100,7) + int(m.group(2)) * pow(100,6) \
            + int(m.group(3)) * pow(100,4) + int(m.group(4))
    return version

def getver_str(ver, rel=True):
    major = ver / pow(100,7)
    middle = (ver - major * pow(100,7)) / pow(100,6)
    minor = (ver - major * pow(100,7) - middle * pow(100,6)) / pow(100,4)
    release = ver - major * pow(100,7) - middle * pow(100,6) - minor * pow(100,4)
    if rel:
        return "%i.%i.%i (%i)" % (major, middle, minor, release)
    return "%i.%i.%i" % (major, middle, minor)

def preInstall(storage, payload, ksdata, instClass):
    # Vstorage - select group if clustername given
    if ksdata.vstorage.clustername and not instClass.is_free:
        payload.selectGroup("vstorage")
        if ksdata.vstorage.client and ksdata.vstorage.shaman:
            payload.selectGroup("high-availability")
# Temporary disabled: need to think whats to do in such case
#        if not ksdata.vstorage.client:
#            map(lambda x: payload.deselectGroup(x), ["vz", "ps", "templates"])

def postInstall(storage, ksdata, instClass):
    try:
        execInSysroot("/usr/libexec/create_bridges.py", [])
    except:
        pass

    product_name = ""
    product_releasef = instClass.get_release_f
    repover = []

    # Detect version ISO
    repover.append(calculate_ver(re.match("^.*(\d+)\.(\d+)\.(\d+) \((\d+)\).*", productVersion, re.I)))

    # Correct release file
    with open(getSysroot() + product_releasef, "r+") as f:
        data = f.read()
        product_name = re.sub(" [0-9]*\.[0-9]*\.[0-9]* \([a-zA-Z0-9\.]*\).*", "", data).rstrip()
        repover.append(calculate_ver(re.match("^.*(\d+)\.(\d+)\.(\d+) \((\d+)\).*", data, re.I)))
        f.seek(0)
        f.write("%s %s\n" % (product_name, getver_str(sorted(repover, reverse=True)[0])))
        f.truncate()
        f.close()

    src = getSysroot() + "/etc/pki/rpm-gpg"
    [ rpm_import(os.path.join(src, i)) for i in os.listdir(src) ]

def create_vstorage_input(pw):
    pipe = os.pipe()
    os.write(pipe[1], pw + '\n')
    os.close(pipe[1])
    return pipe[0]

@property
def check_vstorage_avail():
    if not os.path.exists(vstorage_bin):
        return False
    return True

def process_vstorage_auth_dirs(cluster_name):
    keyfile_dir = "/etc/vstorage/clusters/%s" % cluster_name
    keyfile = os.path.join(keyfile_dir, "auth_digest.key")

    if not os.path.exists(keyfile_dir):
        os.makedirs(keyfile_dir)

    if os.path.exists(keyfile):
        os.unlink(keyfile)

def vstorage_auth_node(cluster_name, vstorage_clusterpw, reg_timeout, stage2=False):
    if stage2 and not check_vstorage_avail:
        log.error("%s absent in 2nd stage" % vstorage_bin)
        return 255

    if stage2:
        rootpath = "/"
        process_vstorage_auth_dirs(cluster_name)
    else:
        rootpath = getSysroot()

    while (reg_timeout):
        input_data = create_vstorage_input(vstorage_clusterpw)
        auth_start_time = time.time()
        rc = execWithRedirect("/usr/bin/vstorage", [ "-vvvvvvvv", "-t", "30", "-c", cluster_name,
            "auth-node", "-P"], root=rootpath, stdin=input_data)
        overal_auth_time = time.time() - auth_start_time
        os.close(input_data)
        log.debug("vstorage auth-node exit code is %i, "
            "working time: %.2f sec." % (rc, overal_auth_time))

        if rc == 0 or rc == 253 or reg_timeout == 1:    #code 253 - wrong password.
            break

        if overal_auth_time > 60:
            log.warning("Time exceeded execution of command "
                "vstorage auth-node: %.2f sec." % overal_auth_time)
        else:
            time.sleep(60 - overal_auth_time)
        reg_timeout -= 1

    if rc == 0:
        log.info("The server has been successfully authenticated.")
    elif rc == 253:
        log.error("Failed to authenticate the server in the cluster: wrong password")
    else:
        log.error("Authentication failed: Could not connect to the cluster! " \
            "see vstorage error(s) in anaconda program log")

    return rc

def checkVstorageLicense(clustername, license):
    if not check_vstorage_avail:
        log.error("%s absent in 2nd stage" % vstorage_bin)
        return ""

    buf = execWithCapture(vstorage_bin, [ "view-license", "-p", license ])

    log.info("Checking entered %s key %s" % (clustername, license))

    if "status=\"VALID\"" in buf or "keyserver_host=" in buf:
        return license

    log.error("Entered key %s is not activation %s key or valid key.:\n%s" % \
            (license, clustername, buf))
    return ""

def get_nonlocal_ip_list(ipv4_only = True):
    ip_list = []

    for ip in network.getIPs():
        if ip == "127.0.0.1" or ip == "::1" or ip.startswith("fe80"):
            continue
        if ipv4_only and ip.find(':') > 0 and ip.find('.') == -1:
            continue
        ip_list.append(ip)

    if len(ip_list) == 0:
        log.warning("No network interface card on this server has "
                "a valid IP address assigned.\nMake sure the "
                "network for this server is configured properly.")

    return ip_list

# Equivalent to standart execWithRedirect, but close stdout after execution
def execWithRedirectCl(*argv, **kwargs):
    rc = execWithRedirect(*argv, **kwargs)
    try:
        kwargs["stdout"].close()
    except:
        pass
    return rc

def inside_vm():
    """Check if the installer is running inside a VM."""

    return hvt_checker.CPU().has_feature("hypervisor")

def vt_supported():
    """
    Check that basic virtualization is enabled and the CPU has full support
    of required features.
    """

    cpu = hvt_checker.CPU.get()
    if cpu is None:
        return False

    return cpu.vt_enabled() and cpu.full_featured() and cpu.kvm_loaded()

# Check that drive fast enough to be SSD
def is_ssd(disk):
    # HACK HACK HACK - provide ssd_hack=sda,sdb,sdg for SSDs
    try:
        f = open("/proc/cmdline", "r")
        for param in f.read().split(" "):
            if not param.startswith("ssd_hack="):
                continue
            for ssd in re.sub("^ssd_hack=", "", param).split(","):
                if ssd.rstrip('\n') == disk.name:
                    return True
        f.close()
    except:
        pass

    if inside_vm():
        return False

    ssd_iops_barrier = 2000

    # Use precached values
    if disk.iops:
        log.info("IOPS for %s already measured: %i\n" % (disk.name, disk.iops))
        if disk.iops > ssd_iops_barrier:
            return True
        return False

    devicename = "/dev/%s" % disk.name

    ioping_out = execWithCapture("/usr/bin/ioping",
        [ "-D",  "-B",  "-q",  "-i",  "0", "-c", "10k", "-w", "1s", devicename ],
        filter_stderr=True)
    ioping = re.match("^[0-9]+ [0-9]+ (\d+) .*$", ioping_out, re.I)
    if ioping and ioping.group(1):
        log.info("Measured IOPS for %s: %s" % (devicename, ioping.group(1)))
        disk.iops = int(ioping.group(1))
        if int(ioping.group(1)) > ssd_iops_barrier:
            log.info("Detected SSD drive: %s" % devicename)
            return True
    else:
        log.info("Failed to measure IOPS for %s" % devicename)

    return False

def minimal_hardware(disks, join_cs):
    if join_cs:
        return not bool([ d for d in disks if d.cap_csdata ])
    return not bool([ d for d in disks if d.cap_vzdata ])

def in_cmdline(cparam):
    result = False
    try:
        f = open("/proc/cmdline", "r")
        for param in f.read().split(" "):
            if param.rstrip('\n') == cparam:
                result = True
                break
        f.close()
    except:
        pass

    return result
