from pyanaconda.i18n import _
import pykickstart.commands as commands
from pykickstart.errors import formatErrorMsg, KickstartError, KickstartValueError
from pyanaconda.iutil import execInSysroot, execWithRedirect, getSysroot, execWithCapture
from pyanaconda import network
from pyanaconda.flags import flags
from pyanaconda.virtuozzo_c import vstorage_auth_node, create_vstorage_input, \
    get_nonlocal_ip_list, execWithRedirectCl, checkVstorageLicense, DoVZAutopart
from netaddr import IPNetwork
from copy import copy
import shutil
from logging import getLogger
log = getLogger("anaconda")
import os
import re
import warnings
import time
from blivet.size import Size

class Vstorage(commands.vstorage.VZ7_Vstorage):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.vstorage.VZ7_Vstorage.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.mds_port = 2510
        self.started_services = []
        self.vstorage_bin = "/usr/bin/vstorage"
        self.common_opts = ["-vvvvvvvv", "-t", "90", "-c" ]
        self.logfile = "/tmp/vstorage.log"
        self.shaman_bin = "/usr/sbin/shaman"
        self.snmp_mib_tree = ".1.3.6.1.4.1.26171"
        self.client_dir = ""
        self.mds_min_size = Size("10GiB")
        # Save 20% of data on SSDs for reserve, backported from PCS6
        self.ssd_reserved_coeff = 0.8
        self.default_vstorage_path = "/vstorage"
        self.vstorage_service = "/usr/libexec/vstorage/vstorage-service"

    def stop_service(self, service):
        rc = execInSysroot("killall", [service])
        if rc:
            log.info("Failed to stop %s service: %i . Looks like already stopped", service, rc)
            return False
        return True

    def start_service(self, service, args):
        if not os.path.exists(getSysroot() + self.vstorage_service):
            log.error("VStorage service binary %s does not exist", self.vstorage_service)
            return False
        rc = execInSysroot(self.vstorage_service, ["start", self.clustername, service] + args)
        if rc:
            log.error("Failed to start %s service: %i", service, rc)
            return False
        self.started_services.append(service)
        return True

    def stop_services(self):
        for service in self.started_services:
            self.stop_service(service)

    def determine_subnet(self, ip):
        """Determine subnet address in CIDR notation.
           Return subnet address in string format or None"""
        net_info = execWithCapture("ip", ["addr", "show"], filter_stderr=True)

        data = re.search('%s[^\s]*' % ip, net_info)
        if not data:
            return None

        return str(IPNetwork(data.group()).cidr)

    def get_vstorage_client_ip(self):
        """Retrieve ip address of client from 'vstorage stat' output.
           Here we assume that this address can be used to contact cluster.
           Return ip address in string format or None"""
        stat = execWithCapture("vstorage", ["-c", self.clustername, "stat", "-X"],
            filter_stderr=True, root=getSysroot())

        cluster_ips = set(re.findall('<host>([^\<host>]*)', stat))
        node_ips = set(get_nonlocal_ip_list())

        client_ip = node_ips & cluster_ips
        if not client_ip:
            return None

        return client_ip.pop()

    def configure_snmpd(self):
        ip = self.get_vstorage_client_ip()
        if not ip:
            log.error("Failed to determine client ip address")
            return

        subnet = self.determine_subnet(ip)
        if not subnet:
            log.error("Failed to retrieve cluster subnet address")
            return

        try:
            f = open(getSysroot() + "/etc/snmp/snmpd.local.conf", "a")

            try:
                sample = "rwcommunity" + "\tparallels" + "\t%s" + "\t%s"

                # grant read/write access for localhost
                print >> f, sample % ("127.0.0.1", self.snmp_mib_tree)
                # grant read/write access for cluster subnet
                print >> f, sample % (subnet, self.snmp_mib_tree)

            finally:
                f.close()

        except IOError:
            log.error("Failed to edit configuration file of the SNMP daemon")

    def prepare_ssd_sizes(self, ssds, clients):
        space_vals = []
        # Sort - more diskspace - first
        ssds = sorted(ssds, key=lambda k: k[1], reverse = True)

        # Variant 1 - less or eq clients then ssds: sort and use max
        if clients <= len(ssds):
            # Set equal sizes for all ssds
            for mp in sorted(set([ d[0] for d in ssds ])):
                sizes = [ d for d in ssds if d[0] == mp ]
                splitted = len(sizes)
                allsize = sum([ d[1] for d in sizes ])
                for ssd in sizes:
                    ssd[1] = allsize/splitted
            for i in range(clients):
                space_vals.append((ssds[i][0], ssds[i][1]))
            return space_vals

        # Variant2 - clients more than SSDs, so need to split smth
        # Will split bigger, then resort and requrse and so one
        ssds[0][1] /= 2
        ssds.append(copy(ssds[0]))

        return self.prepare_ssd_sizes(ssds, clients)

    def get_ssd_caches(self, storage):
        ssd_cache_parts = []
        for mp in sorted([ mp for mp in storage.mountpoints if
            mp.startswith("/mnt/ssd") ]):
            # Calculate size
            try:
                statfs = os.statvfs(getSysroot() + mp)
                size = int(statfs.f_bsize * statfs.f_bavail / Size("1MiB"))
                ssd_cache_parts.append([ mp, int(size * self.ssd_reserved_coeff) ])
                log.info("Detected SSD cache mounted at %s size as %s MiB" % (mp, size))
            except Exception, e:
                log.error("Can't detect available ssd cache size for %s\nError: %s" % (mp, str(e)))

        if not ssd_cache_parts:
            return []

        clients = 0
        if self.need_mds_create():
            # Reserve size on 1st drive for MDS
            ssd_cache_parts[0][1] -= self.mds_min_size/Size("1MiB")
        if self.cs_join and self.cs_ssd:
            clients += len(self.vstorage_parts(storage))
        if self.client and self.client_ssd:
            clients += 1

        return self.prepare_ssd_sizes(ssd_cache_parts, clients)

    def vstorage_parts(self, storage):
        return [ p for p in storage.partitions + storage.lvs +
            storage.mdarrays if p.format.type == "ext4cs" ]

    def execute(self, storage, ksdata, instClass):
        if not self.clustername or instClass.is_free:
            return True

        if not os.path.exists(getSysroot() + self.vstorage_bin):
            log.error("%s absent in installed system" % self.vstorage_bin)
            return False

        if not self.need_mds_create():
            vstorage_auth_node(self.clustername, self.clusterpw, self.reg_timeout + 1)

        # Calculate sizes for SSD caches
        ssd_caches = self.get_ssd_caches(storage)

        # Check for main directory
        if not os.path.exists(getSysroot() + self.default_vstorage_path):
            # And create it if does not exists
            os.mkdir(getSysroot() + self.default_vstorage_path)

        # Create or Join MDS
        if self.need_mds():
            # Create MDS on SSD if it available
            mds_path = "/mnt/ssd1"
            if not os.path.exists(getSysroot() + mds_path):
                mds_path = "%s/%s-mds" % (self.default_vstorage_path, self.clustername)
                if not os.path.exists(getSysroot() + mds_path):
                    os.mkdir(getSysroot() + mds_path)
            mds_path += "/data"
            stdin=None
            parg = []
            initarg = []
            if self.need_mds_create():
                stdin=create_vstorage_input(self.clusterpw)
                parg = ["-P"]
                initarg = ["--init-cluster"]
            rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                "make-mds" ] + initarg + [ "-a", "%s:%i" % (self.mds_ip, self.mds_port),
                "-r", mds_path ] + parg, root=getSysroot(), stdin=stdin,
                stdout = open(self.logfile, "a+"))
            if rc:
                log.error("MDS %s failed: %i" % ("create" if self.need_mds_create() else "join", rc))
                return False
            if self.need_mds_create():
                if not self.start_service("mdsd", [mds_path]):
                    return False
                # Create Container and VM dirs on clean cluster
                tmpdir = "/tmp/vstorage_m"
                os.mkdir(getSysroot() + tmpdir)
                execWithRedirectCl("/bin/mount", ["-t", "fuse.vstorage", "vstorage://%s" %
                    self.clustername , tmpdir], stdout = open(self.logfile, "a+"),
                    root=getSysroot())
                os.mkdir(getSysroot() + tmpdir + "/private")
                os.mkdir(getSysroot() + tmpdir + "/vmprivate")
                execWithRedirectCl("/bin/umount", [tmpdir], stdout = open(self.logfile, "a+"),
                    root=getSysroot())
                shutil.rmtree(getSysroot() + tmpdir)

        # Create CSes
        if self.cs_join:
            csnum = 0
            for part in self.vstorage_parts(storage):
                reg_timeout = self.reg_timeout + 1
                cspath = part.format.mountpoint + "/data"
                while (reg_timeout):
                    ssd_params = []
                    if len(ssd_caches) and self.cs_ssd:
                        ssh_cache_path = ssd_caches[0][0] + "/%s-cs%i-ssd" % (self.clustername, csnum)
                        if not os.path.exists(getSysroot() + ssh_cache_path):
                            os.mkdir(getSysroot() + ssh_cache_path)
                        ssd_params.append("--journal=%s" % ssh_cache_path)
                        ssd_params.append("--journal-size=%i" % ssd_caches[0][1])
                        ssd_caches.pop(0)
                    start_time = time.time()
                    rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                        "make-cs", "-r", cspath, "--tier=%s" % part.format.get_tier() ] + ssd_params,
                        stdout = open(self.logfile, "a+"), root=getSysroot())
                    overal_time = time.time() - start_time
                    if rc:
                        log.error("CS join failed: %i, will try again", rc)
                        reg_timeout -= 1
                        if overal_time > 60:
                            log.warning("Time exceeded execution of command %s make-cs" % self.vstorage_bin)
                            continue
                        time.sleep(60 - overal_time)
                        continue
                    if not self.start_service("csd", [cspath]):
                        log.warning("Failed to start CSD service for %s mpoint" % cspath)
                    break
                if rc:
                    log.error("CS join failed: %i", rc)
                csnum += 1

        # Create Client
        if self.client:
            # Create client fstab and dir
            client_cache_param = []
            if len(ssd_caches) and self.client_ssd:
                client_cache_dir = "%s/client_cache" % ssd_caches[0][0]
                if not os.path.exists(getSysroot() + client_cache_dir):
                    os.mkdir(getSysroot() + client_cache_dir)
                client_cache_param.append("cache=%s" % client_cache_dir)
                client_cache_param.append("cachesize=%i" % ssd_caches[0][1])
            self.client_dir = "%s/%s" % (self.default_vstorage_path, self.clustername)
            if not os.path.exists(getSysroot() + self.client_dir):
                os.mkdir(getSysroot() + self.client_dir)
            f = open("%s/etc/fstab" % getSysroot(), 'a+')
            client_cache_param.append("_netdev")
            f.write('\n' + "vstorage://%s %s fuse.vstorage %s 0 0" % (self.clustername,
                self.client_dir, ','.join(client_cache_param)) + '\n')
            f.close()
            os.symlink(self.client_dir + "/private", getSysroot() + "/vz/private")
            os.symlink(self.client_dir + "/vmprivate", getSysroot() + "/vz/vmprivate")

        # Check vstorage license
        if self.key:
            self.key = checkVstorageLicense(self.clustername, self.key)

        # Vstorage key
        if self.key:
            rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                "load-license", "-p", self.key], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to activate the %s license key: %s" % \
                     (self.clustername, self.key))
                self.key = ""

        # Enable High Availability
        if self.shaman and self.client and self.key:
            rc = execWithRedirectCl("/bin/mount", [self.client_dir], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to mount %s Client to %s : %i" % (self.productname, self.client_dir, rc))
                return False
            rc = execWithRedirectCl(self.shaman_bin, [ "-c", self.clustername, "join", "-r" ],
                stdout = open(self.logfile, "a+"), root=getSysroot())
            if rc:
                log.error("Shaman join failed: %i...", rc)
                return False
            rc = execWithRedirectCl("/bin/umount", [self.client_dir], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to umount %s Client to %s : %i" % (self.productname, self.client_dir, rc))

            self.configure_snmpd()

        # Stop
        self.stop_services()

        return True

class VzttUrlmap(commands.vztturlmap.VZ7_Vztturlmap):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.vztturlmap.VZ7_Vztturlmap.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.url_map_f = "/vz/template/conf/vztt/url.map"
        self.url_map_l = "/etc/vztt/url.map"

    def execute(self, storage, ksdata, instClass):
        log.info("Changing %s..." % self.url_map_l)
        urlmapf = getSysroot() + self.url_map_f
        if not os.path.exists(urlmapf):
            return False
        f = open(urlmapf).read()
        for replace_strings_array in self.urls:
            if re.search(r'\\=', replace_strings_array):
                replace_strings = replace_strings_array.split('\=')
            else:
                replace_strings = replace_strings_array.split('=')
            if len(replace_strings) == 2:
                replace_what_re = '^' + re.sub( r'\$', '#?\$', replace_strings[0]) + '.*'
                fnew = ""
                line_not_found = True
                for url_map_line in f.splitlines():
                    url_map_line_unmodifyed = url_map_line
                    url_map_line = re.sub(replace_what_re, replace_strings[0] + "\t" +  replace_strings[1], url_map_line_unmodifyed)
                    if url_map_line_unmodifyed != url_map_line:
                        line_not_found = False
                    fnew += "%s\n" % url_map_line
                if line_not_found:
                    fnew += "%s\n" % (replace_strings[0] + "\t" +  replace_strings[1])
                f = fnew
        open(urlmapf, 'w').write(f)
        return True

class Cep(commands.cep.VZ7_CEP):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.cep.VZ7_CEP.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.command = "/usr/bin/prlsrvctl set --cep "
        self.set_enabled = self.command + "on"
        self.set_disabled = self.command + "off"

    def execute(self, storage, ksdata, instClass):
        if self.enabled:
            log.info("Enabling CEP...")
            instClass.prl_disp_first_start([self.set_enabled])
            return
        log.info("CEP is disabled.")
        instClass.prl_disp_first_start([self.set_disabled])

class Key(commands.key.RHEL5_Key):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.key.RHEL5_Key.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.logfile = "/tmp/license.log"
        self.vzlicload = "/usr/sbin/vzlicload"
        self.vzlicupdate = "/usr/sbin/vzlicupdate"
        self.vzlicview = "/usr/sbin/vzlicview"

    def checkLicense(self):
        if not self.key:
            return False

        if not os.access(getSysroot() + self.vzlicview, os.X_OK):
            log.error("An error occurred while installing the license! " \
                "%s is not found or not accessible" % self.vzlicview)
            return False

        buf = execWithCapture(self.vzlicview, ["-p", self.key ], root=getSysroot(),
            filter_stderr=True)
        log.info("Checking entered key %s" % self.key)
        if buf.lower().find("keyserver_host=") != -1:
            type_activation = True
            log.info("Entered key %s is an activation key." % self.key)
        elif buf.lower().find("status=\"valid\"") != -1:
            type_activation = False
            log.info("Entered key %s is a valid product key." % self.key)
        else:
            log.error("Entered key %s is not activation key or valid product key." % self.key)
            return False

        return True

    def processLicense(self):
        if not self.checkLicense():
            return

        if not os.access(getSysroot() +  self.vzlicload, os.X_OK):
            log.error("An error occurred while installing the license! " \
                "%s is not found or not accessible" % self.vzlicload)
            return

        log.info("Installing the new license...")
        if execWithRedirectCl(self.vzlicload, ["-p", self.key],
                 stdout = open(self.logfile, "a+"),
                 root=getSysroot()):
            log.error("An error occurred while installing the license. "
                                  "For details, see %s ." % os.path.basename(self.logfile))
            return

        log.info("The license has been installed successfully!")

    def execute(self, storage, ksdata, instClass):
        if instClass.is_free:
            return

        if self.key and not self.skip:
            self.processLicense()

class Prlnet(commands.prlnet.VZ7_Prlnet):
    def execute(self, storage, ksdata, instClass):
        if not self.used:
            return
        if not self.name:
            self.name = "Host-Only"
        log.info("Set IP addresses scope for network {self.name}, " \
            "start IP {self.ip_scope_start}, end IP {self.ip_scope_end} " \
            "ip/mask {self.ip} DHCP IP {self.dhcp_ip} " \
            "start IP6 {self.ip6_scope_start}, end IP6 {self.ip6_scope_end} " \
            "ip6/mask {self.ip6} DHCP IP6 {self.dhcp_ip6} " \
            "dhcp-server: {self.dhcp_server} dhcp6-server: {self.dhcp6_server}".format(**locals()))
        opts = ["net", "set", self.name]
        if self.ip_scope_start:
             opts.extend(["--ip-scope-start", self.ip_scope_start, \
                          "--ip-scope-end", self.ip_scope_end])
        if self.ip6_scope_start:
             opts.extend(["--ip6-scope-start", self.ip6_scope_start, \
                          "--ip6-scope-end", self.ip6_scope_end])
        if self.ip:
            opts.extend(["--ip", self.ip])
        if self.ip6:
            opts.extend(["--ip6", self.ip6])

        if self.dhcp_ip:
            opts.extend(["--dhcp-ip", self.dhcp_ip])
        if self.dhcp_ip6:
            opts.extend(["--dhcp-ip6", self.dhcp_ip6])

        if self.dhcp_server:
            opts.extend(["--dhcp-server", self.dhcp_server])
        if self.dhcp6_server:
            opts.extend(["--dhcp6-server", self.dhcp6_server])

        instClass.prl_disp_first_start(["/usr/bin/prlsrvctl " + " ".join(opts)])

class AutoPart(commands.autopart.RHEL7_AutoPart):
    def __init__(self, writePriority=100, *args, **kwargs):
        super(AutoPart, self).__init__(writePriority=writePriority, *args, **kwargs)

    def parse(self, args):
        retval = super(AutoPart, self).parse(args)

        ignored_keys = []
        if self.fstype:
            ignored_keys.append("--fstype")
        if self.encrypted:
            ignored_keys.append("--encrypted")
        if self.type:
            ignored_keys.append("--type")
        if not self.lvm:
            ignored_keys.append("--no-lvm")
        if ignored_keys:
            warnings.warn("The autopart kickstart key does not support: %s ." % \
                ", ".join(ignored_keys))

        return retval

    def execute(self, storage, ksdata, instClass):
        if not self.autopart:
            return

        DoVZAutopart(instClass, storage).execute()

class Readykernel(commands.readykernel.VZ7_Readykernel):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.readykernel.VZ7_Readykernel.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.command = "/usr/sbin/readykernel"
        self.autoupdate_cmd = ["autoupdate"]
        self.autoupdate_enable = self.autoupdate_cmd + ["enable"]
        self.autoupdate_disable = self.autoupdate_cmd + ["disable"]

    def execute(self, storage, ksdata, instClass):
        if instClass.is_free:
            return

        # Call utility
        rc = execInSysroot(self.command, self.autoupdate_enable if self.autoupdate else self.autoupdate_disable)
        if rc:
            log.error("Failed to execute %s command: %i", self.command, rc)
            return False
        return True
